package org.processmining.humbertogarciacaballero.models.gui.controllers;

import java.awt.event.MouseEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.swing.SwingUtilities;

import org.processmining.humbertogarciacaballero.models.controller.PNetVisualizerController;
import org.processmining.humbertogarciacaballero.models.model.PNConstants;

import prefuse.Display;
import prefuse.controls.DragControl;
import prefuse.data.Graph;
import prefuse.data.expression.ColumnExpression;
import prefuse.data.expression.ComparisonPredicate;
import prefuse.data.expression.NumericLiteral;
import prefuse.visual.EdgeItem;
import prefuse.visual.NodeItem;
import prefuse.visual.VisualItem;

public class CustomDragControl extends DragControl {

    private final PNetVisualizerController controller;
    private final Point2D from;
    private final Point2D to;
    
    private boolean left;
    private final Point2D auxPoint;
    private final Point2D auxFrom;
    private final ComparisonPredicate predicate;
    private final Set<NodeItem> set;
    
    /**
     * Creates a new drag control that invokes an action upon drag events.
     * @param action the action to run when drag events occur.
     */
    public CustomDragControl(String action) {
    	super(action);
    	controller = PNetVisualizerController.getInstance();
    	from = new Point2D.Double();
    	to = new Point2D.Double();
    	left = false;
    	auxPoint = new Point2D.Double();
    	auxFrom = new Point2D.Double();
    	predicate = new ComparisonPredicate(ComparisonPredicate.EQ, new ColumnExpression(PNConstants.NODE_LAYER_KEY), new NumericLiteral(-1));
    	set = new HashSet<>();
    }
    
    /**
     * @see prefuse.controls.Control#itemPressed(prefuse.visual.VisualItem, java.awt.event.MouseEvent)
     */
    public void itemPressed(VisualItem item, MouseEvent e) {
        if (!SwingUtilities.isLeftMouseButton(e) || !(item instanceof NodeItem)) return;
        from.setLocation(item.getX(), item.getY());
        auxPoint.setLocation(item.getX(), item.getY());
        auxFrom.setLocation(item.getX(), item.getY());
        left = false;
        set.clear();
        super.itemPressed(item, e);
    }
    
    public void itemReleased(VisualItem item, MouseEvent e) {
        if (!SwingUtilities.isLeftMouseButton(e) || !(item instanceof NodeItem)) return;
        if ( dragged ) {
            to.setLocation(item.getX(), item.getY());
            controller.dragNode(item, from, to);
        }
        set.clear();
        super.itemReleased(item, e);
    }
    
    /**
     * @see prefuse.controls.Control#itemDragged(prefuse.visual.VisualItem, java.awt.event.MouseEvent)
     */
    public void itemDragged(VisualItem item, MouseEvent e) {
        if (!SwingUtilities.isLeftMouseButton(e) || !(item instanceof NodeItem)) return;
        dragged = true;
        Display d = (Display)e.getComponent();
        d.getAbsoluteCoordinate(e.getPoint(), temp);
        double dx = temp.getX()-down.getX();
        double dy = temp.getY()-down.getY();
        double x = item.getX();
        double y = item.getY();

        item.setStartX(x);  item.setStartY(y);
        item.setX(x+dx);    item.setY(y+dy);
        item.setEndX(x+dx); item.setEndY(y+dy);

        AffineTransform transform = new AffineTransform();
        transform.translate(dx, dy);
        
        updateDataStructures(item);
        updatePorts((NodeItem) item);
        
        if ( repaint )
            item.getVisualization().repaint();
        
        down.setLocation(temp);
        if ( action != null )
            d.getVisualization().run(action);
    }
    
    private void updateDataStructures(VisualItem item) {
        double currentPosition, lastPosition;
        
        switch (controller.getLayoutOrientation()) {
        case TOP_TO_BOTTOM:
        case BOTTOM_TO_TOP:
            currentPosition = item.getX();
            lastPosition = auxPoint.getX();
            break;
        default:
            currentPosition = item.getY();
            lastPosition = auxPoint.getY();
            break;
        }
        
        if (currentPosition < lastPosition) {
            if (!left) {
                left = true;
                set.clear();
                auxFrom.setLocation(auxPoint);
            }
        } else {
            if (left) {
                left = false;
                set.clear();
                auxFrom.setLocation(auxPoint);
            }
        }
        
        auxPoint.setLocation(item.getX(), item.getY());
    }
    
    private void updatePorts(NodeItem nodeItem) {
        Graph graph = nodeItem.getGraph();
        int layer = nodeItem.getInt(PNConstants.NODE_LAYER_KEY);
        Set<NodeItem> neighbors = new HashSet<>();
        Iterator<NodeItem> iter = nodeItem.neighbors();
        while (iter.hasNext()) {
            NodeItem neighbor = iter.next();
            if (neighbor.getInt(PNConstants.NODE_LAYER_KEY) == layer + 1 || neighbor.getInt(PNConstants.NODE_LAYER_KEY) == layer - 1) {
                neighbors.add(neighbor);
            }
        }

        predicate.setRightExpression(new NumericLiteral(layer));
        iter = nodeItem.getGraph().getNodes().tuples(predicate);
        while (iter.hasNext()) {
            NodeItem neighbor = iter.next();
            if (isCrossedNode(nodeItem, neighbor) && !set.contains(neighbor)) {
                Iterator<NodeItem> iter2 = neighbor.neighbors();
                while (iter2.hasNext()) {
                    NodeItem neighbor2 = iter2.next();
                    if (neighbors.contains(neighbor2)) {
                        EdgeItem edgeItem = (EdgeItem) graph.getEdge(nodeItem, neighbor2);
                        boolean target = true;
                        if (edgeItem == null) {
                            edgeItem = (EdgeItem) graph.getEdge(neighbor2, nodeItem);
                            target = false;
                        }
                        
                        EdgeItem edgeItem2 = (EdgeItem) graph.getEdge(neighbor, neighbor2);
                        boolean target2 = true;
                        if (edgeItem2 == null) {
                            edgeItem2 = (EdgeItem) graph.getEdge(neighbor2, neighbor);
                            target2 = false;
                        }
                        
                        float aux;
                        if (target) {
                            if (target2) {
                                aux = edgeItem.getFloat(PNConstants.EDGE_TARGET_PORT_OFFSET);
                                edgeItem.setFloat(PNConstants.EDGE_TARGET_PORT_OFFSET, edgeItem2.getFloat(PNConstants.EDGE_TARGET_PORT_OFFSET));
                                edgeItem2.setFloat(PNConstants.EDGE_TARGET_PORT_OFFSET, aux);
                            } else {
                                aux = edgeItem.getFloat(PNConstants.EDGE_TARGET_PORT_OFFSET);
                                edgeItem.setFloat(PNConstants.EDGE_TARGET_PORT_OFFSET, edgeItem2.getFloat(PNConstants.EDGE_SOURCE_PORT_OFFSET));
                                edgeItem2.setFloat(PNConstants.EDGE_SOURCE_PORT_OFFSET, aux);                                
                            }
                        } else {
                            if (target2) {
                                aux = edgeItem.getFloat(PNConstants.EDGE_SOURCE_PORT_OFFSET);
                                edgeItem.setFloat(PNConstants.EDGE_SOURCE_PORT_OFFSET, edgeItem2.getFloat(PNConstants.EDGE_TARGET_PORT_OFFSET));
                                edgeItem2.setFloat(PNConstants.EDGE_TARGET_PORT_OFFSET, aux);
                            } else {
                                aux = edgeItem.getFloat(PNConstants.EDGE_SOURCE_PORT_OFFSET);
                                edgeItem.setFloat(PNConstants.EDGE_SOURCE_PORT_OFFSET, edgeItem2.getFloat(PNConstants.EDGE_SOURCE_PORT_OFFSET));
                                edgeItem2.setFloat(PNConstants.EDGE_SOURCE_PORT_OFFSET, aux);                                
                            }
                        }
                    }
                }
                
                set.add(neighbor);
            }
        }
    }

    private boolean isCrossedNode(VisualItem nodeItem, VisualItem neighbor) {
        double nodeItemPosition, neighborPosition, fromPosition;
        
        switch (controller.getLayoutOrientation()) {
        case TOP_TO_BOTTOM:
        case BOTTOM_TO_TOP:
            nodeItemPosition = nodeItem.getX();
            neighborPosition = neighbor.getX();
            fromPosition = auxFrom.getX();
            break;
        default:
            nodeItemPosition = nodeItem.getY();
            neighborPosition = neighbor.getY();
            fromPosition = auxFrom.getY();
            break;
        }
        
        if (nodeItemPosition < fromPosition) {
            if (neighborPosition < fromPosition) {
                return nodeItemPosition < neighborPosition;
            }
        } else {
            if (neighborPosition > fromPosition) {
                return nodeItemPosition > neighborPosition;
            }
        }
        
        return false;
    }
}
