package org.processmining.humbertogarciacaballero.models.model.commands;

import java.util.LinkedList;
import java.util.Observable;
import java.util.Observer;

public class CommandManager extends Observable implements Observer {

    private LinkedList<AbstractCommand> historyList;
    private LinkedList<AbstractCommand> redoList;

    public enum CommandManagerState {
        EMPTY(false, false), FULL(true, true), ONLY_REDO(true, false), ONLY_UNDO(false, true);
        private boolean redo;
        private boolean undo;

        private CommandManagerState(boolean redo, boolean undo) {
            this.redo = redo;
            this.undo = undo;
        }

        public boolean isRedo() {
            return redo;
        }

        public boolean isUndo() {
            return undo;
        }
    }

    public CommandManager() {
        historyList = new LinkedList<>();
        redoList = new LinkedList<>();
    }

    public CommandManagerState executeCommand(AbstractCommand command) {
        if (command instanceof Undo) {
            undo();
            return getState();
        }
        if (command instanceof Redo) {
            redo();
            return getState();
        }

        command.addObserver(this);
        command.doIt();
        historyList.addFirst(command);
        if (redoList.size() > 0) {
            redoList.clear();
        }
        setChanged();
        notifyObservers(getState());
        return getState();
    }

    public void undo() {
        if (!historyList.isEmpty()) {
            AbstractCommand undoCommand = historyList.removeFirst();
            undoCommand.addObserver(this);
            try {
                undoCommand.undoIt();
            } catch (Exception e) {
                e.printStackTrace();
            }
            redoList.addFirst(undoCommand);
            setChanged();
            notifyObservers(getState());
        }
    }

    public void redo() {
        if (!redoList.isEmpty()) {
            AbstractCommand redoCommand = redoList.removeFirst();
            redoCommand.addObserver(this);
            try {
                redoCommand.doIt();
            } catch (Exception e) {
                e.printStackTrace();
            }
            historyList.addFirst(redoCommand);
            setChanged();
            notifyObservers(getState());
        }
    }

    public boolean isThereMoreUndo() {
        return !historyList.isEmpty();
    }

    public boolean isThereMoreRedo() {
        return !redoList.isEmpty();
    }

    public CommandManagerState getState() {
        if (historyList.isEmpty() && redoList.isEmpty())
            return CommandManagerState.EMPTY;
        if (historyList.isEmpty() && !redoList.isEmpty())
            return CommandManagerState.ONLY_REDO;
        if (!historyList.isEmpty() && redoList.isEmpty())
            return CommandManagerState.ONLY_UNDO;
        return CommandManagerState.FULL;
    }

    @Override
    public void update(Observable o, Object arg) {
        o.deleteObserver(this);
        setChanged();
        notifyObservers();
    }
}
